/* Copyright 2004, University of Colorado */

/*
 * CVS Info -
 * Filename : $Source: /cvsroot/phet/phetcommon/src/edu/colorado/phet/common/view/ControlPanel.java,v $
 * Branch : $Name:  $
 * Modified by : $Author: ronlemaster $
 * Revision : $Revision: 1.42 $
 * Date modified : $Date: 2006/02/03 20:18:47 $
 */
package edu.colorado.phet.common.view;

import edu.colorado.phet.common.application.Module;

import javax.swing.*;
import java.awt.*;

/**
 * ControlPanel is the panel that contains the controls for the simulation.
 *
 * @author Ron LeMaster
 * @version $Revision: 1.42 $
 */
public class ControlPanel extends JPanel {

    private ContentPanel contentPanel; // holds the controls
    private JScrollPane scrollPane;
    private Scrollable scrollPolicy;

    /**
     * Constructs a ControlPanel for the specified module.
     *
     * @param module
     * @deprecated
     */
    public ControlPanel( Module module ) {
        this();
    }

    /**
     * Constructor
     */
    public ControlPanel() {
        setLayout( new BorderLayout() );
        contentPanel = new ContentPanel();
        contentPanel.setFillNone();
        scrollPolicy = new DefaultScrollPolicy( contentPanel );
        scrollPane = new JScrollPane( contentPanel, JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED,
                                      JScrollPane.HORIZONTAL_SCROLLBAR_NEVER );
        add( scrollPane, BorderLayout.CENTER );

        // Macintosh fix
        {
            scrollPane.setOpaque( false );
            scrollPane.getViewport().setOpaque( false );
        }
    }

    /**
     * Gets the content panel, which holds the controls.
     *
     * @return the content panel
     */
    public ContentPanel getContentPanel() {
        return contentPanel;
    }

    /**
     * Adds a component to the control area.
     *
     * @param component
     * @return the component argument
     */
    public Component addControl( Component component ) {
        return contentPanel.add( component );
    }

    /**
     * Adds a component to the content panel, to fill the width.
     *
     * @param component
     */
    public void addControlFullWidth( Component component ) {
        contentPanel.addFullWidth( component );
    }

    /**
     * For backward compatibility.
     *
     * @param component
     * @return the component argument
     * @deprecated
     */
    public Component add( Component component ) {
        return addControl( component );
    }

    /**
     * For backward compatibility.
     *
     * @param component
     * @deprecated
     */
    public void addFullWidth( Component component ) {
        addControlFullWidth( component );
    }

    /**
     * Removes a component from the control area
     *
     * @param comp
     */
    public void removeControl( Component comp ) {
        contentPanel.remove( comp );
        contentPanel.invalidate();
        contentPanel.validate();
    }

    /**
     * Adds a full-width separator to the control panel.
     */
    public void addSeparator() {
        addControlFullWidth( new JSeparator() );
    }

    /**
     * Adds vertical space to the control panel using a vertical strut.
     *
     * @param space the amount of space, in pixels
     */
    public void addVerticalSpace( int space ) {
        if( space > 0 ) {
            JPanel spacePanel = new JPanel();
            spacePanel.setLayout( new BoxLayout( spacePanel, BoxLayout.Y_AXIS ) );
            spacePanel.add( Box.createVerticalStrut( space ) );
            addControlFullWidth( spacePanel );
        }
    }

    /**
     * Sets the insets used in the content panel.
     *
     * @param insets
     */
    public void setInsets( Insets insets ) {
        contentPanel.setInsets( insets );
    }

    /**
     * Sets a new scroll policy for the JComponent in this ControlPanel.
     *
     * @param scrollPolicy
     */
    public void setScrollPolicy( Scrollable scrollPolicy ) {
        this.scrollPolicy = scrollPolicy;
    }

    private int getVisibleScrollBarWidth() {
        return isScrollBarVisible() ? getScrollBarWidth() : 0;
    }

    private int getScrollBarWidth() {
        return scrollPane.getVerticalScrollBar().getPreferredSize().width;
    }

    private boolean isScrollBarVisible() {
        return scrollPane.getVerticalScrollBar().isVisible();
    }

    /**
     * The DefaultScrollPolicy uses the preferred size of the component, adding on the width of the visible scroll bars
     * in the ControlPanel.
     */
    public class DefaultScrollPolicy implements Scrollable {
        JComponent component;

        public DefaultScrollPolicy( JComponent component ) {
            this.component = component;
        }

        public boolean getScrollableTracksViewportHeight() {
            return false;
        }

        public boolean getScrollableTracksViewportWidth() {
            return false;
        }

        public Dimension getPreferredScrollableViewportSize() {
            return new Dimension( component.getPreferredSize().width + getVisibleScrollBarWidth(), component.getPreferredSize().height );
        }

        public int getScrollableBlockIncrement( Rectangle visibleRect, int orientation, int direction ) {
            return 10;
        }

        public int getScrollableUnitIncrement( Rectangle visibleRect, int orientation, int direction ) {
            return 10;
        }
    }

    /**
     * This is where the controls go in the ControlPanel.  It delegates its Scrollable interface to
     */
    public class ContentPanel extends VerticalLayoutPanel implements Scrollable {

        public boolean getScrollableTracksViewportHeight() {
            return scrollPolicy.getScrollableTracksViewportHeight();
        }

        public boolean getScrollableTracksViewportWidth() {
            return scrollPolicy.getScrollableTracksViewportWidth();
        }

        public Dimension getPreferredScrollableViewportSize() {
            return scrollPolicy.getPreferredScrollableViewportSize();
        }

        public int getScrollableBlockIncrement( Rectangle visibleRect, int orientation, int direction ) {
            return scrollPolicy.getScrollableBlockIncrement( visibleRect, orientation, direction );
        }

        public int getScrollableUnitIncrement( Rectangle visibleRect, int orientation, int direction ) {
            return scrollPolicy.getScrollableUnitIncrement( visibleRect, orientation, direction );
        }
    }
}